from __future__ import absolute_import, print_function, division

# Load configuration
from . import config
del config

from .core import YarnCluster

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



from __future__ import absolute_import, division, print_function

import sys

PY2 = sys.version_info.major == 2

if PY2:
    from urlparse import urlparse
    from backports import weakref
else:
    import weakref  # noqa
    from urllib.parse import urlparse  # noqa



from __future__ import division, print_function, absolute_import

import os
import shutil
import time

import dask
from dask.distributed import Client
from distributed.utils_test import inc
import pytest
import skein

from dask_yarn import YarnCluster
from dask_yarn.core import _make_specification, _make_submit_specification
from .conftest import check_is_shutdown

try:
    import bokeh  # noqa
    bokeh_installed = True
except Exception:
    bokeh_installed = False


@pytest.mark.parametrize('deploy_mode', ['remote', 'local'])
def test_basic(deploy_mode, skein_client, conda_env):
    with YarnCluster(environment=conda_env,
                     deploy_mode=deploy_mode,
                     worker_memory='512 MiB',
                     scheduler_memory='512 MiB',
                     name='test-basic',
                     skein_client=skein_client) as cluster:
        # Smoketest repr
        repr(cluster)

        if bokeh_installed:
            assert cluster.dashboard_link is not None

        # Scale up
        cluster.scale(2)

        with Client(cluster) as client:
            future = client.submit(inc, 10)
            assert future.result() == 11
            client.get_versions(check=True)

        # Check that 2 workers exist
        start = time.time()
        while len(cluster.workers()) != 2:
            time.sleep(0.1)
            assert time.time() < start + 5, "timeout cluster.scale(2)"

        # Scale down
        cluster.scale(1)

        start = time.time()
        while len(cluster.workers()) != 1:
            time.sleep(0.1)
            assert time.time() < start + 5, "timeout cluster.scale(1)"

    check_is_shutdown(skein_client, cluster.app_id)


def test_from_specification(skein_client, conda_env, tmpdir):
    spec = _make_specification(environment=conda_env,
                               worker_memory='512 MiB',
                               scheduler_memory='512 MiB',
                               name='dask-yarn-test-from-specification')
    fn = os.path.join(str(tmpdir), 'spec.yaml')
    with open(fn, 'w') as f:
        f.write(spec.to_yaml())

    with YarnCluster.from_specification(fn, skein_client=skein_client) as cluster:
        with Client(cluster):
            pass

    check_is_shutdown(skein_client, cluster.app_id)


def test_from_specification_errors():
    bad_spec = skein.ApplicationSpec.from_yaml("""
        name: bad_spec
        services:
          bad:
            resources:
              memory: 1 GiB
              vcores: 1
            commands:
              - exit 1
        """)
    with pytest.raises(ValueError):
        YarnCluster.from_specification(bad_spec)

    with pytest.raises(TypeError):
        YarnCluster.from_specification(object())


def test_from_application_id(skein_client, conda_env):
    with YarnCluster(environment=conda_env,
                     worker_memory='512 MiB',
                     scheduler_memory='512 MiB',
                     name='test-from-application-id',
                     skein_client=skein_client) as cluster:

        # Connect to the application with the application id
        cluster2 = YarnCluster.from_application_id(cluster.app_id, skein_client)

        cluster2.scale(1)

        start = time.time()
        while len(cluster2.workers()) != 1:
            time.sleep(0.1)
            assert time.time() < start + 5, "timeout cluster.scale(1)"

        del cluster2

        # Cluster is still running, finalizer not run in cluster2
        assert len(cluster.workers()) == 1

    check_is_shutdown(skein_client, cluster.app_id)


def test_from_current(skein_client, conda_env, monkeypatch, tmpdir):
    # Not running in a container
    with pytest.raises(ValueError) as exc:
        YarnCluster.from_current()
    assert str(exc.value) == "Not running inside a container"

    with YarnCluster(environment=conda_env,
                     worker_memory='512 MiB',
                     scheduler_memory='512 MiB',
                     name='test-from-current',
                     skein_client=skein_client) as cluster:

        # Patch environment so it looks like a container
        container_id = 'container_1526134340424_0012_01_000005'
        cont_dir = tmpdir.mkdir(container_id)
        shutil.copyfile(skein_client.security.cert_path,
                        str(cont_dir.join(".skein.crt")))
        shutil.copyfile(skein_client.security.key_path,
                        str(cont_dir.join(".skein.pem")))

        for key, val in [('SKEIN_APPLICATION_ID', cluster.app_id),
                         ('CONTAINER_ID', container_id),
                         ('SKEIN_APPMASTER_ADDRESS', cluster.application_client.address),
                         ('LOCAL_DIRS', str(tmpdir))]:
            monkeypatch.setenv(key, val)

        import skein.core
        monkeypatch.setattr(skein.core, 'properties', skein.core.Properties())

        cluster2 = YarnCluster.from_current()
        assert cluster2.app_id == cluster.app_id
        assert cluster2.scheduler_address == cluster.scheduler_address

        # Smoketest method
        cluster2.scale(1)

        start = time.time()
        while len(cluster2.workers()) != 1:
            time.sleep(0.1)
            assert time.time() < start + 5, "timeout cluster.scale(1)"

        del cluster2

        # Cluster is still running, finalizer not run in cluster2
        assert len(cluster.workers()) == 1

    check_is_shutdown(skein_client, cluster.app_id)


def test_configuration():
    config = {'yarn': {
        'environment': 'myenv.tar.gz',
        'queue': 'myqueue',
        'name': 'dask-yarn-tests',
        'tags': ['a', 'b', 'c'],
        'specification': None,
        'worker': {'memory': '1234 MiB', 'count': 1, 'vcores': 1, 'restarts': -1,
                   'env': {'foo': 'bar'}},
        'scheduler': {'memory': '1234 MiB', 'vcores': 1}}
    }

    with dask.config.set(config):
        spec = _make_specification()
        assert spec.name == 'dask-yarn-tests'
        assert spec.queue == 'myqueue'
        assert spec.tags == {'a', 'b', 'c'}
        assert spec.services['dask.worker'].resources.memory == 1234
        assert spec.services['dask.worker'].env == {'foo': 'bar'}
        assert spec.services['dask.scheduler'].resources.memory == 1234


def test_configuration_full_specification(conda_env, tmpdir):
    spec = _make_specification(environment=conda_env,
                               worker_memory='512 MiB',
                               scheduler_memory='512 MiB',
                               name='test-configuration-full-specification')
    fn = os.path.join(str(tmpdir), 'spec.yaml')
    with open(fn, 'w') as f:
        f.write(spec.to_yaml())

    # path to full specification
    with dask.config.set({'yarn': {'specification': fn}}):
        spec = _make_specification()
        assert spec == spec

    # full specification inlined
    with dask.config.set({'yarn': {'specification': spec.to_dict()}}):
        spec = _make_specification()
        assert spec == spec

    # when overrides specified, ignores full specification
    override = {'specification': spec.to_dict()}
    override.update(dask.config.get('yarn'))
    override['name'] = 'config-name'
    with dask.config.set({'yarn': override}):
        spec = _make_specification(environment='foo',
                                   name='test-name',
                                   n_workers=4)
        assert spec.services['dask.worker'].instances == 4
        assert spec.name == 'test-name'


def test_make_specification_errors():
    with dask.config.set({'yarn.environment': None}):
        with pytest.raises(ValueError) as info:
            _make_specification()
        assert 'conda-pack' in str(info.value)

    with pytest.raises(ValueError) as info:
        _make_specification(environment='foo.tar.gz', deploy_mode='unknown')
    assert 'deploy_mode' in str(info.value)

    with pytest.raises(ValueError) as info:
        _make_specification(environment='foo.tar.gz', worker_memory=1234)
    assert '1234 MiB' in str(info.value)

    with pytest.raises(ValueError) as info:
        _make_specification(environment='foo.tar.gz', scheduler_memory=1234)
    assert '1234 MiB' in str(info.value)


@pytest.mark.parametrize('deploy_mode', ['local', 'remote'])
def test_make_submit_specification(deploy_mode):
    spec = _make_submit_specification('../script.py',
                                      deploy_mode=deploy_mode,
                                      environment='myenv.tar.gz',
                                      name='test-name',
                                      client_vcores=2,
                                      client_memory='2 GiB')

    assert spec.name == 'test-name'
    if deploy_mode == 'local':
        assert set(spec.services) == {'dask.worker'}
        assert 'environment' in spec.services['dask.worker'].files
    else:
        client = spec.services['dask.client']
        scheduler = spec.services['dask.scheduler']
        assert client.files['environment'] == scheduler.files['environment']
        assert client.files['script.py'].source.startswith('file:///')
        assert client.resources.memory == 2048
        assert client.resources.vcores == 2

    config = {
        'yarn.name': 'dask-yarn-tests',
        'yarn.queue': 'myqueue',
        'yarn.environment': 'myenv.tar.gz',
        'yarn.client.memory': '1234 MiB',
        'yarn.client.vcores': 2,
        'yarn.client.env': {'foo': 'bar'},
    }

    with dask.config.set(config):
        spec = _make_submit_specification('script.py', deploy_mode=deploy_mode)
        assert spec.name == 'dask-yarn-tests'
        assert spec.queue == 'myqueue'
        if deploy_mode == 'remote':
            assert spec.services['dask.client'].resources.memory == 1234
            assert spec.services['dask.client'].resources.vcores == 2
            assert spec.services['dask.client'].env == {'foo': 'bar'}


def test_environment_relative_paths(conda_env):
    a = _make_specification(environment=conda_env).to_dict()
    b = _make_specification(environment=os.path.relpath(conda_env)).to_dict()
    assert a == b



from __future__ import division, print_function, absolute_import

import pytest

import dask_yarn
from dask_yarn.cli import main
from .conftest import ensure_shutdown, wait_for_completion, get_logs


def run_command(command, error=False):
    with pytest.raises(SystemExit) as exc:
        main([arg for arg in command.split(' ') if arg])
    if error:
        assert exc.value.code != 0
    else:
        assert exc.value.code == 0


@pytest.mark.parametrize('command',
                         ['',
                          'submit',
                          'status',
                          'kill',
                          'services',
                          'services scheduler',
                          'services worker',
                          'services client'])
def test_cli_help(command, capfd):
    run_command(command + ' -h')

    out, err = capfd.readouterr()
    assert not err
    assert 'usage: dask-yarn' in out


@pytest.mark.parametrize('group', ['', 'services'])
def test_cli_call_command_group(group, capfd):
    run_command(group, error=True)

    out, err = capfd.readouterr()
    assert not out
    assert 'usage: dask-yarn' in err


def test_cli_version(capfd):
    run_command('--version')

    out, err = capfd.readouterr()
    assert not err
    assert dask_yarn.__version__ in out


GOOD_TEST_SCRIPT = """
import dask_yarn
from dask.distributed import Client

print("Connecting to cluster")
cluster = dask_yarn.YarnCluster.from_current()

print("Creating Dask Client")
client = Client(cluster)

print("Doing some computation")
assert client.submit(lambda x: x + 1, 1).result() == 2

print("Done!")
"""


BAD_TEST_SCRIPT = """
import dask_yarn
from dask.distributed import Client

print("Connecting to cluster")
cluster = dask_yarn.YarnCluster.from_current()

print("Creating Dask Client")
client = Client(cluster)

print("Intentionally Erroring")
assert False, "Failed!"
"""


scripts = {'good': GOOD_TEST_SCRIPT,
           'bad': BAD_TEST_SCRIPT}


@pytest.mark.parametrize('script_kind, final_status, searchtxt',
                         [('good', 'SUCCEEDED', 'Done!'),
                          ('bad', 'FAILED', 'Failed!')])
def test_cli_submit_and_status(script_kind, final_status, searchtxt,
                               tmpdir, conda_env, skein_client, capfd):
    script_path = str(tmpdir.join('script.py'))
    with open(script_path, 'w') as fil:
        fil.write(scripts[script_kind])

    run_command('submit '
                '--name test-cli-submit-and-status '
                '--environment %s '
                '--worker-count 1 '
                '--worker-memory 256MiB '
                '--worker-vcores 1 '
                '--scheduler-memory 256MiB '
                '--scheduler-vcores 1 '
                '--client-memory 128MiB '
                '--client-vcores 1 '
                '%s' % (conda_env, script_path))
    out, err = capfd.readouterr()
    # Logs go to err
    assert 'INFO' in err
    app_id = out.strip()
    assert '\n' not in app_id

    with ensure_shutdown(skein_client, app_id, status=final_status):
        # Wait for app to start
        skein_client.connect(app_id)

        # `dask-yarn status`
        run_command('status %s' % app_id)
        out, err = capfd.readouterr()
        assert 'INFO' in err
        assert len(out.splitlines()) == 2
        assert 'RUNNING' in out

        wait_for_completion(skein_client, app_id, timeout=60)

    logs = get_logs(app_id)
    assert searchtxt in logs


def test_cli_submit_with_args(tmpdir, conda_env, skein_client, capfd):
    script = ('import sys\n'
              'args = sys.argv[1:]\n'
              'assert args == ["a", "b", "c"]\n'
              'print("Done!")')

    script_path = str(tmpdir.join('script.py'))
    with open(script_path, 'w') as fil:
        fil.write(script)

    run_command('submit '
                '--name test-cli-submit-and-status '
                '--environment %s '
                '--worker-count 0 '
                '--scheduler-memory 256MiB '
                '--scheduler-vcores 1 '
                '--client-memory 128MiB '
                '--client-vcores 1 '
                '%s a b c' % (conda_env, script_path))
    out, err = capfd.readouterr()
    # Logs go to err
    assert 'INFO' in err
    app_id = out.strip()
    assert '\n' not in app_id

    with ensure_shutdown(skein_client, app_id, status='SUCCEEDED'):
        # Wait for app to start
        skein_client.connect(app_id)
        wait_for_completion(skein_client, app_id, timeout=60)

    logs = get_logs(app_id)
    assert 'Done!' in logs


@pytest.mark.parametrize('script_kind, final_status, searchtxt',
                         [('good', 'SUCCEEDED', 'Done!'),
                          ('bad', 'FAILED', 'Failed!')])
def test_cli_submit_local(script_kind, final_status, searchtxt,
                          tmpdir, conda_env, skein_client, capfd):
    if script_kind == 'good':
        error = False
        script = GOOD_TEST_SCRIPT
    else:
        error = True
        script = BAD_TEST_SCRIPT

    script_path = str(tmpdir.join('script.py'))
    with open(script_path, 'w') as fil:
        fil.write(script)

    run_command('submit '
                '--name test-cli-submit-and-status '
                '--environment %s '
                '--deploy-mode local '
                '--worker-count 1 '
                '--worker-memory 256MiB '
                '--worker-vcores 1 '
                '--scheduler-memory 256MiB '
                '--scheduler-vcores 1 '
                '--client-memory 128MiB '
                '--client-vcores 1 '
                '%s' % (conda_env, script_path), error=error)
    out, err = capfd.readouterr()
    # Logs go to err
    assert 'INFO' in err
    assert searchtxt in out or searchtxt in err


def test_cli_kill(tmpdir, conda_env, skein_client, capfd):
    script_path = str(tmpdir.join('script.py'))
    with open(script_path, 'w') as fil:
        fil.write(GOOD_TEST_SCRIPT)

    run_command('submit '
                '--name test-cli-kill '
                '--environment %s '
                '--worker-count 1 '
                '--worker-memory 256MiB '
                '--worker-vcores 1 '
                '--scheduler-memory 256MiB '
                '--scheduler-vcores 1 '
                '--client-memory 128MiB '
                '--client-vcores 1 '
                '%s' % (conda_env, script_path))
    out, err = capfd.readouterr()
    # Logs go to err
    assert 'INFO' in err
    app_id = out.strip()
    assert '\n' not in app_id

    with ensure_shutdown(skein_client, app_id, status='KILLED'):
        # Wait for app to start
        skein_client.connect(app_id)

        # `dask-yarn kill`
        run_command('kill %s' % app_id)
        out, err = capfd.readouterr()
        assert 'INFO' in err
        assert not out

        wait_for_completion(skein_client, app_id, timeout=60)



from __future__ import print_function, division, absolute_import

import argparse
import os
import subprocess
import sys

import skein
from skein.utils import format_table, humanize_timedelta
from tornado import gen
from tornado.ioloop import IOLoop, TimeoutError
from distributed import Scheduler, Nanny
from distributed.utils import ignoring
from distributed.cli.utils import install_signal_handlers, uri_from_host_port
from distributed.proctitle import (enable_proctitle_on_children,
                                   enable_proctitle_on_current)

from . import __version__
from .compat import urlparse
from .core import _make_submit_specification, YarnCluster, _get_skein_client


class _Formatter(argparse.HelpFormatter):
    """Format with a fixed argument width, due to bug in argparse measuring
    argument widths"""
    @property
    def _action_max_length(self):
        return 16

    @_action_max_length.setter
    def _action_max_length(self, value):
        pass

    def _format_args(self, action, default_metavar):
        """Format remainder arguments nicer"""
        get_metavar = self._metavar_formatter(action, default_metavar)
        if action.nargs == argparse.REMAINDER:
            return '[%s...]' % get_metavar(1)
        return super(_Formatter, self)._format_args(action, default_metavar)


class _VersionAction(argparse.Action):
    def __init__(self, option_strings, version=None, dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS, help="Show version then exit"):
        super(_VersionAction, self).__init__(option_strings=option_strings,
                                             dest=dest, default=default,
                                             nargs=0, help=help)
        self.version = version

    def __call__(self, parser, namespace, values, option_string=None):
        print(self.version % {'prog': parser.prog})
        sys.exit(0)


def fail(msg):
    print(msg, file=sys.stderr)
    sys.exit(1)


def add_help(parser):
    parser.add_argument("--help", "-h", action='help',
                        help="Show this help message then exit")


def arg(*args, **kwargs):
    return (args, kwargs)


def subcommand(subparsers, name, help, *args):
    def _(func):
        parser = subparsers.add_parser(name,
                                       help=help,
                                       formatter_class=_Formatter,
                                       description=help,
                                       add_help=False)
        parser.set_defaults(func=func)
        for arg in args:
            parser.add_argument(*arg[0], **arg[1])
        add_help(parser)
        func.parser = parser
        return func
    return _


def node(subs, name, help):
    @subcommand(subs, name, help)
    def f():
        fail(f.parser.format_usage())
    f.subs = f.parser.add_subparsers(metavar='command', dest='command')
    f.subs.required = True
    return f


entry = argparse.ArgumentParser(prog="dask-yarn",
                                description="Deploy Dask on Apache YARN",
                                formatter_class=_Formatter,
                                add_help=False)
add_help(entry)
entry.add_argument("--version", action=_VersionAction,
                   version='%(prog)s ' + __version__,
                   help="Show version then exit")
entry.set_defaults(func=lambda: fail(entry.format_usage()))
entry_subs = entry.add_subparsers(metavar='command', dest='command')
entry_subs.required = True


def _parse_env(service, env):
    out = {}
    if env is None:
        return out
    for item in env:
        elements = item.split('=')
        if len(elements) != 2:
            raise ValueError("Invalid parameter to --%s-env: %r" % (service, env))
        key, val = elements
        out[key.strip()] = val.strip()
    return out


# Exposed for testing
def _parse_submit_kwargs(**kwargs):
    if kwargs.get('worker_env') is not None:
        kwargs['worker_env'] = _parse_env('worker', kwargs['worker_env'])
    if kwargs.get('client_env') is not None:
        kwargs['client_env'] = _parse_env('client', kwargs['client_env'])
    if kwargs.get('tags') is not None:
        kwargs['tags'] = set(map(str.strip, kwargs["tags"].split(",")))
    if kwargs.get('worker_count') is not None:
        kwargs['n_workers'] = kwargs.pop('worker_count')
    return kwargs


@subcommand(entry_subs,
            'submit', 'Submit a Dask application to a YARN cluster',
            arg("script", help="Path to a python script to run on the client"),
            arg("args", nargs=argparse.REMAINDER,
                help="Any additional arguments to forward to `script`"),
            arg("--name", help="The application name"),
            arg("--queue", help="The queue to deploy to"),
            arg("--tags",
                help=("A comma-separated list of strings to use as "
                      "tags for this application.")),
            arg("--environment",
                help=("Path to an archived Python environment (either "
                      "``tar.gz`` or ``zip``).")),
            arg("--deploy-mode",
                help=("Either 'remote' (default) or 'local'. If 'remote', the "
                      "scheduler and client will be deployed in a YARN "
                      "container. If 'local', they will be run locally.")),
            arg("--worker-count", type=int,
                help="The number of workers to initially start."),
            arg("--worker-vcores", type=int,
                help="The number of virtual cores to allocate per worker."),
            arg("--worker-memory", type=str,
                help=("The amount of memory to allocate per worker. Accepts a "
                      "unit suffix (e.g. '2 GiB' or '4096 MiB'). Will be "
                      "rounded up to the nearest MiB.")),
            arg("--worker-restarts", type=int,
                help=("The maximum number of worker restarts to allow before "
                      "failing the application. Default is unlimited.")),
            arg("--worker-env", type=str, action='append',
                help=("Environment variables to set on the workers. Pass a "
                      "key-value pair like ``--worker-env key=val``. May "
                      "be used more than once.")),
            arg("--client-vcores", type=int,
                help="The number of virtual cores to allocate for the client."),
            arg("--client-memory", type=str,
                help=("The amount of memory to allocate for the client. "
                      "Accepts a unit suffix (e.g. '2 GiB' or '4096 MiB'). "
                      "Will be rounded up to the nearest MiB.")),
            arg("--client-env", type=str, action='append',
                help=("Environment variables to set on the client. Pass a "
                      "key-value pair like ``--client-env key=val``. May "
                      "be used more than once.")),
            arg("--scheduler-vcores", type=int,
                help="The number of virtual cores to allocate for the scheduler."),
            arg("--scheduler-memory", type=str,
                help=("The amount of memory to allocate for the scheduler. "
                      "Accepts a unit suffix (e.g. '2 GiB' or '4096 MiB'). "
                      "Will be rounded up to the nearest MiB.")))
def submit(script, args=None, **kwargs):
    kwargs = _parse_submit_kwargs(**kwargs)
    args = args or []
    spec = _make_submit_specification(script, args=args, **kwargs)

    if 'dask.scheduler' in spec.services:
        # deploy_mode == 'remote'
        app_id = _get_skein_client().submit(spec)
        print(app_id)
    else:
        # deploy_mode == 'local'
        if not os.path.exists(script):
            raise ValueError("%r doesn't exist locally" % script)

        with YarnCluster.from_specification(spec) as cluster:
            env = dict(os.environ)
            env.update({'DASK_APPLICATION_ID': cluster.app_id,
                        'DASK_APPMASTER_ADDRESS': cluster.application_client.address})

            retcode = subprocess.call([sys.executable, script] + args, env=env)

            if retcode == 0:
                cluster.shutdown("SUCCEEDED")
            else:
                cluster.shutdown("FAILED",
                                 "Exception in submitted dask application, "
                                 "see logs for more details")
                sys.exit(retcode)


app_id = arg('app_id', help='The application id', metavar='APP_ID')


@subcommand(entry_subs,
            'status', 'Check the status of a submitted Dask application',
            app_id)
def status(app_id):
    report = _get_skein_client().application_report(app_id)
    header = ['application_id',
              'name',
              'state',
              'status',
              'containers',
              'vcores',
              'memory',
              'runtime']
    data = [(report.id,
             report.name,
             report.state,
             report.final_status,
             report.usage.num_used_containers,
             report.usage.used_resources.vcores,
             report.usage.used_resources.memory,
             humanize_timedelta(report.runtime))]
    print(format_table(header, data))


@subcommand(entry_subs,
            'kill', 'Kill a Dask application',
            app_id)
def kill(app_id):
    _get_skein_client().kill_application(app_id)


services = node(entry_subs, 'services', 'Manage Dask services')


@subcommand(services.subs,
            'scheduler', 'Start a Dask scheduler process')
def scheduler():  # pragma: nocover
    app_client = skein.ApplicationClient.from_current()

    enable_proctitle_on_current()
    enable_proctitle_on_children()

    if sys.platform.startswith('linux'):
        import resource   # module fails importing on Windows
        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        limit = max(soft, hard // 2)
        resource.setrlimit(resource.RLIMIT_NOFILE, (limit, hard))

    addr = uri_from_host_port('', None, 0)

    loop = IOLoop.current()

    services = {}
    bokeh = False
    with ignoring(ImportError):
        from distributed.bokeh.scheduler import BokehScheduler
        services[('bokeh', 0)] = (BokehScheduler, {})
        bokeh = True

    scheduler = Scheduler(loop=loop, services=services)
    scheduler.start(addr)

    install_signal_handlers(loop)

    # Set dask.dashboard before dask.scheduler since the YarnCluster object
    # waits on dask.scheduler only
    if bokeh:
        bokeh_port = scheduler.services['bokeh'].port
        bokeh_host = urlparse(scheduler.address).hostname
        bokeh_address = 'http://%s:%d' % (bokeh_host, bokeh_port)

        app_client.kv['dask.dashboard'] = bokeh_address.encode()

    app_client.kv['dask.scheduler'] = scheduler.address.encode()

    try:
        loop.start()
        loop.close()
    finally:
        scheduler.stop()


@subcommand(services.subs,
            'worker', 'Start a Dask worker process',
            arg("--nthreads", type=int,
                help=("Number of threads. Defaults to number of vcores in "
                      "container")),
            arg("--memory_limit", type=str,
                help=("Maximum memory available to the worker. This can be an "
                      "integer (in bytes), a string (like '5 GiB' or '500 "
                      "MiB'), or 0 (no memory management). Defaults to the "
                      "container memory limit.")))
def worker(nthreads=None, memory_limit=None):  # pragma: nocover
    enable_proctitle_on_current()
    enable_proctitle_on_children()

    if memory_limit is None:
        memory_limit = int(skein.properties.container_resources.memory * 2**20)
    if nthreads is None:
        nthreads = skein.properties.container_resources.vcores

    app_client = skein.ApplicationClient.from_current()

    scheduler = app_client.kv.wait('dask.scheduler').decode()

    loop = IOLoop.current()

    worker = Nanny(scheduler, ncores=nthreads, loop=loop,
                   memory_limit=memory_limit, worker_port=0)

    @gen.coroutine
    def close(signalnum):
        worker._close(timeout=2)

    install_signal_handlers(loop, cleanup=close)

    @gen.coroutine
    def run():
        yield worker._start(None)
        while worker.status != 'closed':
            yield gen.sleep(0.2)

    try:
        loop.run_sync(run)
    except (KeyboardInterrupt, TimeoutError):
        pass


@subcommand(services.subs,
            'client', 'Start a Dask client process',
            arg("script", help="Path to a Python script to run."),
            arg("args", nargs=argparse.REMAINDER,
                help="Any additional arguments to forward to `script`"))
def client(script, args=None):  # pragma: nocover
    app = skein.ApplicationClient.from_current()
    args = args or []

    if not os.path.exists(script):
        raise ValueError("%r doesn't exist" % script)

    retcode = subprocess.call([sys.executable, script] + args)

    if retcode == 0:
        app.shutdown("SUCCEEDED")
    else:
        print("User submitted application %s failed with returncode "
              "%d, shutting down." % (script, retcode))
        app.shutdown("FAILED",
                     "Exception in submitted dask application, "
                     "see logs for more details")


def main(args=None):
    kwargs = vars(entry.parse_args(args=args))
    kwargs.pop('command', None)  # Drop unnecessary `command` arg
    func = kwargs.pop('func')
    func(**kwargs)
    sys.exit(0)


if __name__ == '__main__':  # pragma: nocover
    main()



from __future__ import absolute_import, print_function, division

import math
import os
import warnings

import dask
from dask.distributed import get_client, LocalCluster
from distributed.utils import format_bytes, PeriodicCallback, log_errors

import skein
from skein.utils import cached_property

from .compat import weakref, urlparse


_memory_error = """Memory specification for the `{0}` take string parameters
with units like "4 GiB" or "2048 MiB"

You provided:       {1}
Perhaps you meant: "{1} MiB"
"""

_one_MiB = 2**20


def parse_memory(memory, field):
    if isinstance(memory, str):
        memory = dask.utils.parse_bytes(memory)
    elif memory < _one_MiB:
        raise ValueError(_memory_error.format(field, memory))
    return int(math.ceil(memory / _one_MiB))


def lookup(kwargs, a, b):
    return kwargs[a] if kwargs.get(a) is not None else dask.config.get(b)


def _get_skein_client(skein_client=None):
    if skein_client is None:
        # Silence warning about credentials not being written yet
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            return skein.Client()
    return skein_client


# exposed for testing only

def _make_specification(**kwargs):
    """ Create specification to run Dask Cluster

    This creates a ``skein.ApplicationSpec`` to run a dask cluster with the
    scheduler in a YARN container. See the docstring for ``YarnCluster`` for
    more details.
    """
    if (all(v is None for v in kwargs.values()) and
            dask.config.get('yarn.specification')):
        # No overrides and full specification in configuration
        spec = dask.config.get('yarn.specification')
        if isinstance(spec, dict):
            return skein.ApplicationSpec.from_dict(spec)
        return skein.ApplicationSpec.from_file(spec)

    deploy_mode = kwargs.get('deploy_mode') or 'remote'
    if deploy_mode not in {'remote', 'local'}:
        raise ValueError("`deploy_mode` must be one of {'remote', 'local'}, "
                         "got %r" % deploy_mode)

    name = lookup(kwargs, 'name', 'yarn.name')
    queue = lookup(kwargs, 'queue', 'yarn.queue')
    tags = lookup(kwargs, 'tags', 'yarn.tags')

    environment = lookup(kwargs, 'environment', 'yarn.environment')
    if environment is None:
        msg = ("You must provide a path to an archived Python environment for "
               "the workers.\n"
               "This is commonly achieved through conda-pack or venv-pack.\n\n"
               "See http://yarn.dask.org/environments.html for more information.")
        raise ValueError(msg)

    n_workers = lookup(kwargs, 'n_workers', 'yarn.worker.count')
    worker_restarts = lookup(kwargs, 'worker_restarts', 'yarn.worker.restarts')
    worker_env = lookup(kwargs, 'worker_env', 'yarn.worker.env')
    worker_vcores = lookup(kwargs, 'worker_vcores', 'yarn.worker.vcores')
    worker_memory = parse_memory(lookup(kwargs, 'worker_memory', 'yarn.worker.memory'),
                                 'worker')

    services = {}

    if deploy_mode == 'remote':
        scheduler_vcores = lookup(kwargs, 'scheduler_vcores',
                                  'yarn.scheduler.vcores')
        scheduler_memory = parse_memory(lookup(kwargs, 'scheduler_memory',
                                               'yarn.scheduler.memory'),
                                        'scheduler')

        services['dask.scheduler'] = skein.Service(
            instances=1,
            resources=skein.Resources(
                vcores=scheduler_vcores,
                memory=scheduler_memory
            ),
            max_restarts=0,
            files={'environment': environment},
            commands=['source environment/bin/activate',
                      'dask-yarn services scheduler']
        )
        worker_depends = ['dask.scheduler']
    else:
        worker_depends = None

    services['dask.worker'] = skein.Service(
        instances=n_workers,
        resources=skein.Resources(
            vcores=worker_vcores,
            memory=worker_memory
        ),
        max_restarts=worker_restarts,
        depends=worker_depends,
        files={'environment': environment},
        env=worker_env,
        commands=['source environment/bin/activate',
                  'dask-yarn services worker']
    )

    spec = skein.ApplicationSpec(name=name,
                                 queue=queue,
                                 tags=tags,
                                 services=services)
    return spec


def _make_submit_specification(script, args=(), **kwargs):
    spec = _make_specification(**kwargs)

    if 'dask.scheduler' in spec.services:
        # deploy_mode == 'remote'
        client_vcores = lookup(kwargs, 'client_vcores', 'yarn.client.vcores')
        client_memory = lookup(kwargs, 'client_memory', 'yarn.client.memory')
        client_env = lookup(kwargs, 'client_env', 'yarn.client.env')
        client_memory = parse_memory(client_memory, 'client')
        environment = spec.services['dask.worker'].files['environment']
        script_name = os.path.basename(script)

        spec.services['dask.client'] = skein.Service(
            instances=1,
            resources=skein.Resources(
                vcores=client_vcores,
                memory=client_memory
            ),
            max_restarts=0,
            depends=['dask.scheduler'],
            files={'environment': environment,
                   script_name: script},
            env=client_env,
            commands=[
                'source environment/bin/activate',
                'dask-yarn services client %s %s' % (script_name, ' '.join(args))
            ]
        )
    return spec


class YarnCluster(object):
    """Start a Dask cluster on YARN.

    You can define default values for this in Dask's ``yarn.yaml``
    configuration file. See http://docs.dask.org/en/latest/configuration.html
    for more information.

    Parameters
    ----------
    environment : str, optional
        Path to an archived Python environment (either ``tar.gz`` or ``zip``).
    n_workers : int, optional
        The number of workers to initially start.
    worker_vcores : int, optional
        The number of virtual cores to allocate per worker.
    worker_memory : str, optional
        The amount of memory to allocate per worker. Accepts a unit suffix
        (e.g. '2 GiB' or '4096 MiB'). Will be rounded up to the nearest MiB.
    worker_restarts : int, optional
        The maximum number of worker restarts to allow before failing the
        application. Default is unlimited.
    worker_env : dict, optional
        A mapping of environment variables to their values. These will be set
        in the worker containers before starting the dask workers.
    scheduler_vcores : int, optional
        The number of virtual cores to allocate per scheduler.
    scheduler_memory : str, optional
        The amount of memory to allocate to the scheduler. Accepts a unit
        suffix (e.g. '2 GiB' or '4096 MiB'). Will be rounded up to the nearest
        MiB.
    deploy_mode : {'remote', 'local'}, optional
        The deploy mode to use. If ``'remote'``, the scheduler will be deployed
        in a YARN container. If ``'local'``, the scheduler will run locally,
        which can be nice for debugging. Default is ``'remote'``.
    name : str, optional
        The application name.
    queue : str, optional
        The queue to deploy to.
    tags : sequence, optional
        A set of strings to use as tags for this application.
    skein_client : skein.Client, optional
        The ``skein.Client`` to use. If not provided, one will be started.

    Examples
    --------
    >>> cluster = YarnCluster(environment='my-env.tar.gz', ...)
    >>> cluster.scale(10)
    """
    def __init__(self,
                 environment=None,
                 n_workers=None,
                 worker_vcores=None,
                 worker_memory=None,
                 worker_restarts=None,
                 worker_env=None,
                 scheduler_vcores=None,
                 scheduler_memory=None,
                 deploy_mode=None,
                 name=None,
                 queue=None,
                 tags=None,
                 skein_client=None):

        spec = _make_specification(environment=environment,
                                   n_workers=n_workers,
                                   worker_vcores=worker_vcores,
                                   worker_memory=worker_memory,
                                   worker_restarts=worker_restarts,
                                   worker_env=worker_env,
                                   scheduler_vcores=scheduler_vcores,
                                   scheduler_memory=scheduler_memory,
                                   deploy_mode=deploy_mode,
                                   name=name,
                                   queue=queue,
                                   tags=tags)

        self._start_cluster(spec, skein_client)

    @cached_property
    def dashboard_link(self):
        """Link to the dask dashboard. None if dashboard isn't running"""
        if self._dashboard_address is None:
            return None
        template = dask.config.get('distributed.dashboard.link')
        dashboard = urlparse(self._dashboard_address)
        params = dict(os.environ)
        params.update({'host': dashboard.hostname, 'port': dashboard.port})
        return template.format(**params)

    @classmethod
    def from_specification(cls, spec, skein_client=None):
        """Start a dask cluster from a skein specification.

        Parameters
        ----------
        spec : skein.ApplicationSpec, dict, or filename
            The application specification to use. Must define at least one
            service: ``'dask.worker'``. If no ``'dask.scheduler'`` service is
            defined, a scheduler will be started locally.
        skein_client : skein.Client, optional
            The ``skein.Client`` to use. If not provided, one will be started.
        """
        self = super(YarnCluster, cls).__new__(cls)
        if isinstance(spec, dict):
            spec = skein.ApplicationSpec.from_dict(spec)
        elif isinstance(spec, str):
            spec = skein.ApplicationSpec.from_file(spec)
        elif not isinstance(spec, skein.ApplicationSpec):
            raise TypeError("spec must be an ApplicationSpec, dict, or path, "
                            "got %r" % type(spec).__name__)
        self._start_cluster(spec, skein_client)
        return self

    def _start_cluster(self, spec, skein_client=None):
        """Start the cluster and initialize state"""

        if 'dask.worker' not in spec.services:
            raise ValueError("Provided Skein specification must include a "
                             "'dask.worker' service")

        skein_client = _get_skein_client(skein_client)

        if 'dask.scheduler' not in spec.services:
            # deploy_mode == 'local'
            self._local_cluster = LocalCluster(n_workers=0,
                                               ip='0.0.0.0',
                                               diagnostics_port=('', 0),
                                               scheduler_port=0)
            scheduler = self._local_cluster.scheduler

            scheduler_address = scheduler.address
            try:
                dashboard_port = scheduler.services['bokeh'].port
            except KeyError:
                dashboard_address = None
            else:
                dashboard_host = urlparse(scheduler_address).hostname
                dashboard_address = 'http://%s:%d' % (dashboard_host, dashboard_port)

            app = skein_client.submit_and_connect(spec)
            try:
                app.kv['dask.scheduler'] = scheduler_address.encode()
                if dashboard_address is not None:
                    app.kv['dask.dashboard'] = dashboard_address.encode()
            except BaseException:
                # Failed to connect, kill the application and reraise
                skein_client.kill_application(app.id)
                raise
        else:
            # deploy_mode == 'remote'
            app = skein_client.submit_and_connect(spec)
            try:
                scheduler_address = app.kv.wait('dask.scheduler').decode()
                dashboard_address = app.kv.get('dask.dashboard')
                if dashboard_address is not None:
                    dashboard_address = dashboard_address.decode()
            except BaseException:
                # Failed to connect, kill the application and reraise
                skein_client.kill_application(app.id)
                raise

        # Ensure application gets cleaned up
        self._finalizer = weakref.finalize(self, app.shutdown)

        self.scheduler_address = scheduler_address
        self._dashboard_address = dashboard_address
        self.app_id = app.id
        self.application_client = app

    @classmethod
    def from_current(cls):
        """Connect to an existing ``YarnCluster`` from inside the cluster.

        Returns
        -------
        YarnCluster
        """
        self = super(YarnCluster, cls).__new__(cls)
        app_id = os.environ.get('DASK_APPLICATION_ID', None)
        app_address = os.environ.get('DASK_APPMASTER_ADDRESS', None)
        if app_id is not None and app_address is not None:
            app = skein.ApplicationClient(app_address, app_id)
        else:
            app = skein.ApplicationClient.from_current()
        self._connect_existing(app)
        return self

    @classmethod
    def from_application_id(cls, app_id, skein_client=None):
        """Connect to an existing ``YarnCluster`` with a given application id.

        Parameters
        ----------
        app_id : str
            The existing cluster's application id.
        skein_client : skein.Client
            The ``skein.Client`` to use. If not provided, one will be started.

        Returns
        -------
        YarnCluster
        """
        self = super(YarnCluster, cls).__new__(cls)
        skein_client = _get_skein_client(skein_client)
        app = skein_client.connect(app_id)
        self._connect_existing(app)
        return self

    def _connect_existing(self, app):
        spec = app.get_specification()
        if 'dask.worker' not in spec.services:
            raise ValueError("%r is not a valid dask cluster" % app.id)

        scheduler_address = app.kv.wait('dask.scheduler').decode()
        dashboard_address = app.kv.get('dask.dashboard')
        if dashboard_address is not None:
            dashboard_address = dashboard_address.decode()

        self.app_id = app.id
        self.application_client = app
        self.scheduler_address = scheduler_address
        self._dashboard_address = dashboard_address
        self._finalizer = None

    def __repr__(self):
        return 'YarnCluster<%s>' % self.app_id

    def _dask_client(self):
        if hasattr(self, '_dask_client_ref'):
            client = self._dask_client_ref()
            if client is not None:
                return client
        client = get_client(address=self.scheduler_address)
        self._dask_client_ref = weakref.ref(client)
        return client

    def shutdown(self, status='SUCCEEDED', diagnostics=None):
        """Shutdown the application.

        Parameters
        ----------
        status : {'SUCCEEDED', 'FAILED', 'KILLED'}, optional
            The yarn application exit status.
        diagnostics : str, optional
            The application exit message, usually used for diagnosing failures.
            Can be seen in the YARN Web UI for completed applications under
            "diagnostics". If not provided, a default will be used.
        """
        if self._finalizer is not None and self._finalizer.peek() is not None:
            self.application_client.shutdown(status=status, diagnostics=diagnostics)
            self._finalizer.detach()  # don't call shutdown later
            # Shutdown in local deploy_mode
            if hasattr(self, '_local_cluster'):
                self._local_cluster.close()
                del self._local_cluster

    def close(self, **kwargs):
        """Close this cluster. An alias for ``shutdown``.

        See Also
        --------
        shutdown
        """
        self.shutdown(**kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def workers(self):
        """A list of all currently running worker containers."""
        return self.application_client.get_containers(services=['dask.worker'])

    def scale_up(self, n, workers=None):
        """Ensure there are atleast n dask workers available for this cluster.

        No-op if ``n`` is less than the current number of workers.

        Examples
        --------
        >>> cluster.scale_up(20)  # ask for twenty workers
        """
        if workers is None:
            workers = self.workers()
        if n > len(workers):
            self.application_client.scale(service='dask.worker', instances=n)

    def scale_down(self, workers):
        """Retire the selected workers.

        Parameters
        ----------
        workers: list
            List of addresses of workers to close.
        """
        self._dask_client().retire_workers(workers)

    def _select_workers_to_close(self, n):
        client = self._dask_client()
        worker_info = client.scheduler_info()['workers']
        # Sort workers by memory used
        workers = sorted((v['metrics']['memory'], k) for k, v in worker_info.items())
        # Return just the ips
        return [w[1] for w in workers[:n]]

    def scale(self, n):
        """Scale cluster to n workers.

        Parameters
        ----------
        n : int
            Target number of workers

        Examples
        --------
        >>> cluster.scale(10)  # scale cluster to ten workers
        """
        workers = self.workers()
        if n >= len(workers):
            return self.scale_up(n, workers=workers)
        else:
            n_to_delete = len(workers) - n
            # Before trying to close running workers, check if there are any
            # pending containers and kill those first.
            pending = [w for w in workers if w.state in ('waiting', 'requested')]

            for c in pending[:n_to_delete]:
                self.application_client.kill_container(c.id)
                n_to_delete -= 1

            if n_to_delete:
                to_close = self._select_workers_to_close(n_to_delete)
                self.scale_down(to_close)

    def _widget_status(self):
        client = self._dask_client()

        workers = client.scheduler_info()['workers']

        n_workers = len(workers)
        cores = sum(w['ncores'] for w in workers.values())
        memory = sum(w['memory_limit'] for w in workers.values())

        text = """
<div>
  <style scoped>
    .dataframe tbody tr th:only-of-type {
        vertical-align: middle;
    }

    .dataframe tbody tr th {
        vertical-align: top;
    }

    .dataframe thead th {
        text-align: right;
    }
  </style>
  <table style="text-align: right;">
    <tr><th>Workers</th> <td>%d</td></tr>
    <tr><th>Cores</th> <td>%d</td></tr>
    <tr><th>Memory</th> <td>%s</td></tr>
  </table>
</div>
""" % (n_workers, cores, format_bytes(memory))
        return text

    def _widget(self):
        """ Create IPython widget for display within a notebook """
        try:
            return self._cached_widget
        except AttributeError:
            pass

        from ipywidgets import Layout, VBox, HBox, IntText, Button, HTML

        client = self._dask_client()

        layout = Layout(width='150px')

        title = HTML('<h2>YarnCluster</h2>')

        status = HTML(self._widget_status(), layout=Layout(min_width='150px'))

        request = IntText(0, description='Workers', layout=layout)
        scale = Button(description='Scale', layout=layout)

        @scale.on_click
        def scale_cb(b):
            with log_errors():
                self.scale(request.value)

        elements = [title,
                    HBox([status, request, scale])]

        if self.dashboard_link is not None:
            link = HTML('<p><b>Dashboard: </b><a href="%s" target="_blank">%s'
                        '</a></p>\n' % (self.dashboard_link, self.dashboard_link))
            elements.append(link)

        self._cached_widget = box = VBox(elements)

        def update():
            status.value = self._widget_status()

        pc = PeriodicCallback(update, 500, io_loop=client.loop)
        pc.start()

        return box

    def _ipython_display_(self, **kwargs):
        try:
            return self._widget()._ipython_display_(**kwargs)
        except ImportError:
            print(self)



from setuptools import setup
import versioneer

with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')

with open('README.rst') as f:
    long_description = f.read()

setup(name='dask-yarn',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      license='BSD',
      description='Deploy dask clusters on Apache YARN',
      long_description=long_description,
      packages=['dask_yarn'],
      include_package_data=True,
      install_requires=install_requires,
      entry_points='''
        [console_scripts]
        dask-yarn=dask_yarn.cli:main
      ''',
      zip_safe=False)



import os
import subprocess
import sys
import time
from contextlib import contextmanager

import skein
import pytest


@pytest.fixture(scope='session')
def conda_env():
    envpath = 'dask-yarn-py%d%d.tar.gz' % sys.version_info[:2]
    if not os.path.exists(envpath):
        conda_pack = pytest.importorskip('conda_pack')
        conda_pack.pack(output=envpath, verbose=True)
    return envpath


@pytest.fixture(scope='session')
def skein_client():
    with skein.Client() as client:
        yield client


def check_is_shutdown(client, app_id, status='SUCCEEDED'):
    timeleft = 5
    report = client.application_report(app_id)
    while report.state not in ('FINISHED', 'FAILED', 'KILLED'):
        time.sleep(0.1)
        timeleft -= 0.1
        if timeleft < 0:
            client.kill_application(app_id)
            assert False, "Application wasn't properly terminated"

    if report.final_status != status:
        logs = get_logs(app_id)
        print('Expected application to terminate with status==%s, got status==%s\n'
              '\n'
              'Application Logs\n'
              '----------------\n'
              '%s' % (status, report.final_status, logs))
        assert report.final_status == status


@contextmanager
def ensure_shutdown(client, app_id, status=None):
    try:
        yield
    except Exception:
        client.kill_application(app_id)
        raise
    else:
        try:
            final_status = wait_for_completion(client, app_id, timeout=5)
        except AssertionError:
            client.kill_application(app_id)
            raise
        assert final_status == status


def wait_for_completion(client, app_id, timeout=30):
    while timeout:
        final_status = client.application_report(app_id).final_status
        if final_status != 'UNDEFINED':
            return final_status
        time.sleep(0.1)
        timeout -= 0.1
    else:
        assert False, "Application timed out"


def get_logs(app_id, tries=3):
    command = ["yarn", "logs", "-applicationId", app_id]
    for i in range(tries - 1):
        try:
            return subprocess.check_output(command).decode()
        except Exception:
            pass
        time.sleep(1)
    return subprocess.check_output(command).decode()
